/* $Id$
 *
 * ebp2prom - Extended Behavior Protocols translator to Promela
 * Copyright (C) 2006-2007  DSRG, Charles University in Prague
 *                          http://dsrg.mff.cuni.cz/
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
 * MA 02110-1301 USA
 *
 */

package org.ow2.dsrg.fm.ebp2promela.generator;

import java.io.*;
import java.util.*;
import java.util.logging.Logger;

import org.ow2.dsrg.fm.bptools.EBPEventTableImpl;
import org.ow2.dsrg.fm.bptools.ProtocolKeeper;
import org.ow2.dsrg.fm.bptools.architecture.PreprocessedArchitecture;
import org.ow2.dsrg.fm.bptools.fsm.CompactFSM;
import org.ow2.dsrg.fm.bptools.fsm.SimpleState;
import org.ow2.dsrg.fm.bptools.fsm.State;
import org.ow2.dsrg.fm.bptools.parser.nodes.BPAbstractComponent;
import org.ow2.dsrg.fm.bptools.parser.nodes.BPComponent;
import org.ow2.dsrg.fm.bptools.parser.nodes.BPTypeDef;
import org.ow2.dsrg.fm.bptools.parser.nodes.BPVarDef;
import org.ow2.dsrg.fm.bptools.parser.nodes.Guard;
import org.ow2.dsrg.fm.bptools.parser.nodes.MethodSignature;
import org.ow2.dsrg.fm.bptools.parser.nodes.MethodSignature.ActualParameter;
import org.ow2.dsrg.fm.bptools.parser.nodes.MethodSignature.Parameter;
import org.ow2.dsrg.fm.bptools.parser.nodes.TrueGuard;
import org.ow2.dsrg.fm.ebp2promela.traversal.Guard2PromelaActions;
import org.ow2.dsrg.fm.ebp2promela.traversal.MultiSyncAnalysisActions;

/**
 * This class exploits the generated javacc parser and dchecker functionality
 * to take a behavior specification in extended behavior protocols and
 * generates somewhat optimized finite automata and data structure in
 * Promela format (with extensive use of embedded c_code).
 *
 * @author Ondrej Sery
 */
public class EmbeddingGenerator {
    
    private static String STATE_LABEL_PREFIX = "_";
    private static String EVENT_REQUEST_PREFIX = "REQ_";
    private static String EVENT_RESPONSE_PREFIX = "RESP_";
    private static String EVENT_MULTISYNC_PREFIX = "M_";
    private static String EVENT_BIT_VAR_PREFIX = "";
    private static String EVENT_COUNTER_VAR_PREFIX = "c";
    private static String SYNC_OFFER_PREFIX = "offer";
    private static String SYNC_UNOFFER_PREFIX = "unoffer";
    private static String FINAL_STATE_SIGNAL = "finish";
    private static String EXIT_STATE_LABEL = "DONE";
    
    /** Creates a new instance of Generator */
    public EmbeddingGenerator(EBPEventTableImpl _et) {
        et = _et;
    }
    
    
    /**
     * Generates output in Promela.
     *
     * @param architecture Preprocessed architecture for Promela generation.
     * @param path Path for generation of files (null for working directory).
     *
     * @throws IOException On problems accessing the output files.
     */
    public void generate(PreprocessedArchitecture architecture, String path)
    throws IOException {
        if (path == null) {
            path = "";
        } else if (path.length() > 0 && !path.endsWith(File.separator)) {
            path = path.concat(File.separator);
        }
        
        
        
        // TODO: By ignoring position of "sync"s among protocols, we slightly
        // change semantic of consent as if the top level consent (from
        // the consent tree) is synced by a union of individual syncs, while
        // other consents are synced by an empty set.
        
        prepareAutomata(architecture);
        
        List<BPComponent> components = architecture.getComponents();
        
        // Multisync analysis
        MultiSyncAnalysisActions multiSyncAnalyzer = new MultiSyncAnalysisActions();
        int i = 0;
        for (BPComponent component : components) {
            multiSyncAnalyzer.analyzeMultiSync(component.getBehavior(), String.valueOf(i));
            i++;
        }
        
        i = 0;
        // TODO: Iterate through the protocols and syncs and generate one
        // component_name.pr file per component
        for (BPComponent component : components) {
            
            // FIXME: dump a file with the automaton
            /*
            try {
                DOTPrintStream s = new DOTPrintStream(
                        path + "automaton_" + i + ".dt");
             
                FSMDOTTraversal t = new FSMDOTTraversal(
                        s,
                        et);
                component.getProtocolKeeper().dotDumpDTFSMS(t);
                s.DOTclose();
                
                DOTPrintStream s2 = new DOTPrintStream(
                        path + "min_automaton_" + i + ".dt");
             
                
                component.getProtocolKeeper().getCompactFSM().printOut(s2,et);
                s2.DOTclose();
                
                
            } catch (Exception e) {
                System.out.println("Could not dump automaton, sorry.");
            }
            */
            Logger.getLogger("EBPTransformation").info(
                    "Generating promela output for component " + i);
            
            /* Old output
            PrintStream pout = new PrintStream(path + "component_" + i + ".pr");
            PrintStream poutData = new PrintStream(path + "component_data_" + i + ".c");
            
            /*  Buffered output */
            PrintStream pout = new PrintStream(
                        new BufferedOutputStream(
                        new FileOutputStream(path + "component_" + i + ".pr"),
                        100000));
            PrintStream poutData = new PrintStream(
                        new BufferedOutputStream(
                        new FileOutputStream(path + "component_data_" + i + ".c"),
                        100000));
            /* */
            generatePreamble(pout, component);
            
            Logger.getLogger("EBPTransformation").info("Preamble generated.");
            
            generateProcess(pout, poutData, component, String.valueOf(i), 
                    multiSyncAnalyzer.getResult());
            
            Logger.getLogger("EBPTransformation").info("Component generated.");
            
            pout.close();
            poutData.close();
            i++;
        }
        
        generateMain(architecture, path, multiSyncAnalyzer.getResult());
        
        Logger.getLogger("EBPTransformation").info(
                "Promela entry point (main.pr) generated.");
        
        generateGlobals(architecture, path, multiSyncAnalyzer.getResult());
        
        Logger.getLogger("EBPTransformation").info(
                "Promela shared data (global.pr) generated.");
        
        InputStream is = ClassLoader.getSystemResourceAsStream("stdebp.pr");
        OutputStream out = new FileOutputStream(path+"stdebp.pr");
        int data;
        while((data=is.read())!=-1){
            out.write(data);
        }
        out.close();
        is.close();
        
        Logger.getLogger("EBPTransformation").info(
                "Set of standard macros (stdebp.pr) copied.");
    }
    
    /**
     * Generates Promela preamble to the output stream. The preamble contains
     * all necessary macros, includes, and so on. The rest is generated using
     * generateProcess().
     */
    private void generatePreamble(PrintStream output, BPComponent component) {
        output.println("#include \"stdebp.pr\"");
        output.println("#include \"global.pr\"");
        output.println();
    }
    
    /**
     * Generates Promela main in file path + main.pr.
     *
     * @throws IOException on output file problems
     */
    private void generateMain(PreprocessedArchitecture architecture,
            String path, Map<Integer, Map<String, Integer>> multiSyncs) throws IOException {
        PrintStream pout = new PrintStream(path + "main.pr");
        List<BPComponent> components = architecture.getComponents();
        
        pout.println("// Standard macros for BP event synchronization.");
        pout.println("#include \"stdebp.pr\"");
        pout.println();
        
        if (!multiSyncs.isEmpty()) {
            pout.println("// Definitions shared by all components.");
            pout.println("c_decl {");
            for (int ev : multiSyncs.keySet()) {
                pout.println("\tchar " + eventToSyncCount(ev) + " = 0;");
            }        
            pout.println("}");
            pout.println();
        }
        
        pout.println("// All components.");
        for (int i = 0; i < components.size(); i++) {
            pout.println("#include \"component_" + i + ".pr\"");
        }
        pout.println();
        
        pout.println(
                "// Main process\n" +
                "init {\n" +
                "\tatomic {\n" +
                "\t\tprocess_count = " + components.size() + ";"
                );
        for (int i = 0; i < components.size(); i++) {
            pout.println("\t\trun c_" + i + "();");
        }
        pout.println("\t\tEBP_INIT_STATS;");
        pout.println("\t}");
        pout.println();
        
        pout.println("\t// Verification stopped. It can be due to \n" +
                "\t// 1) Successfull verification\n" +
                "\t// 2) Bad activity\n" +
                "\t// 3) No activity");
        pout.println("\ttimeout;");
        pout.println("\td_step {\n" +
                "\t\tif\n" +
                "\t\t:: bad_activity ->\n" +
                "\t\t\tif");
        
        // Verbous Bad Activity
        List<Set<Integer>> syncs = architecture.getSyncs();
        Set<Integer> synced = new HashSet<Integer>();
        Set<Integer> unbound = architecture.getUnbound();
        
        for (Set<Integer> sync : syncs) {
            for (Integer event : sync) {
                if (!et.isCommunication(event) || !et.isAccept(event))
                    continue;
                
                if (!synced.contains(event)) {
                    synced.add(event);
                    
                    pout.println("\t\t\t::event==" + eventToBitVar(event) + 
                            " -> c_code{fprintf(stderr,\" *** BAD ACTIVITY: " +
                            "%s\\n\",\"" + eventToStr(event) + "\");}");
                }
            }
        }
        
        
        if (unbound != null) {
            for (Integer event : unbound) {
                if (!et.isCommunication(event) || !et.isAccept(event))
                    continue;
                
                if (!synced.contains(event)) {
                    synced.add(event);
                    
                    pout.println("\t\t\t::event==" + eventToBitVar(event) + 
                            " -> c_code{fprintf(stderr,\" *** BAD ACTIVITY: " +
                            "%s\\n\",\"" + eventToStr(event) + "\");}");
                }
            }
        }
        
        
        pout.println(
                "\t\t\tfi;\n"+
                "\t\t\tassert(0);\n" +
                "\t\t:: else -> skip\n" +
                "\t\tfi;\n" +
                "\t\tend_of_run = 1;\n" +
                "\t}");
        pout.println();
        
        pout.println("\ttimeout;\n" +
                "\td_step {\n" +
                "\t\tif\n" +
                "\t\t:: process_count > 0 ->\n" +
                "\t\t\tc_code{fprintf(stderr,\" *** NO ACTIVITY\\n\");};\n" +
                "\t\t\tassert(0);\n" +
                "\t\t:: else -> skip\n" +
                "\t\tfi;\n" +
                "\t}\n" +
                "}");
        
        pout.close();
    }
    
    /**
     * Generates Promela global shared definitions in file path + global.pr.
     *
     * @throws IOException on output file problems
     */
    private void generateGlobals(PreprocessedArchitecture architecture,
            String path, Map<Integer, Map<String, Integer>> multiSyncs) throws 
            IOException {
        PrintStream pout = new PrintStream(path + "global.pr");
        
        pout.println("#ifndef GLOBAL_PR");
        pout.println("#define GLOBAL_PR");
        pout.println();
        
        List<Set<Integer>> syncs = architecture.getSyncs();
        Set<Integer> synced = new HashSet<Integer>();
        
        for (Set<Integer> sync : syncs) {
            for (Integer event : sync) {
                if (!et.isCommunication(event)) {
                    assert false;
                    continue;
                }
                
                int canonEvent = et.isAccept(event) ? event :
                    et.getOpposite(event);
                
                if (!synced.contains(canonEvent)) {
                    synced.add(canonEvent);
                    
                    pout.println("// EVENT: " + eventToStr(canonEvent));
                    
                    pout.println("#define " + eventToBitVar(
                            canonEvent) + " " + String.valueOf(canonEvent));
                }
            }
        }
        
        Set<Integer> unbound = architecture.getUnbound();
        
        if (unbound != null) {
            for (Integer event : unbound) {
                if (!et.isCommunication(event)) {
                    assert false;
                    continue;
                }
                
                int canonEvent = et.isAccept(event) ? event :
                    et.getOpposite(event);
                
                if (!synced.contains(canonEvent)) {
                    synced.add(canonEvent);
                    
                    pout.println("// UNBOUND EVENT: " + eventToStr(canonEvent));
                    
                    pout.println("#define " + eventToBitVar(
                            canonEvent) + " " + String.valueOf(canonEvent));
                }
            }
        }
        
        Set<Integer> msynced = multiSyncs.keySet();
        
        for (int ms : msynced) {
            if (!et.isMultisync(ms)) {
                assert false;
                continue;
            }

            pout.println("// EVENT: " + eventToStr(ms));

            pout.println("#define " + eventToBitVar(
                    ms) + " " + String.valueOf(ms));
        }
        
        pout.println();
        pout.println("#endif");
        
        pout.close();
    }
    
    /**
     * Generates Promela process to the output stream. Preamble using
     * generatePreamble() should be created before the process.
     */
    private void generateProcess(PrintStream output,PrintStream dataOutput,
            BPComponent component, String id, 
            Map<Integer,Map<String,Integer>> multiSyncs) {
        ProtocolKeeper protocol = component.getProtocolKeeper();
        CompactFSM cfsm =protocol.getCompactFSM();
        SimpleState[] states = cfsm.getStateTable();
        int startState = cfsm.getStartState();
        
        boolean evs[] = new boolean[et.numberOfEvents()];
        boolean ass[] = new boolean[states.length];
        int nevs[] = new int[et.numberOfEvents()];
        Set<Integer> assEvents = new HashSet<Integer>();
        
        for (int i = 0; i < states.length; i++) {
            SimpleState s = states[i];
            int ev = s.nextEdge(State.BEFORE_FIRST_EDGE);
            while (ev != State.AFTER_LAST_EDGE) {
                evs[ev] = true;
                if (et.isCommunication(ev)) {
                    // Nothing special to do
                } else if (et.getAssignment(ev) != null) {
                    assEvents.add(ev);
                    ass[i] = true;
                } else if (et.isLambda(ev)) {
                    // Nothing special to do
                } else if (et.isMultisync(ev)) {
                    // Nothing special to do
                } else  {
                    // Unsuported event type encountered
                    assert false;
                }
                ev = s.nextEdge(ev);
            }
        }
        int ii = 0;
        for (int i = 0; i < et.numberOfEvents(); i++) {
            if (evs[i]) {
                nevs[i] = ii++;
            } else {
                nevs[i] = -1;
            }
        }
        
        dataOutput.println();
        dataOutput.println("// Transition relation");
        dataOutput.println("int tr"+id+"["+states.length+"]["+ii+"]={{");
        boolean first = true;
        for (int i = 0; i < states.length; i++) {
            if (first)
                first = false;
            else
                dataOutput.print("},{");
            generateState(dataOutput, protocol, i, states, id, evs);
        }
        dataOutput.println("}};");
        
        dataOutput.println();
        
        // Component types and local variables
        generateTypes(dataOutput, component, id);
        generateVariables(dataOutput, component, id);
        
        dataOutput.println();
        
        dataOutput.println("// Guard functions");
        int counter = 0;
        for (int i = 0; i < states.length; i++) {
            counter = generateGuardFunctions(dataOutput, protocol, i, states,
                    id, evs, counter);
        }
        
        dataOutput.println("// Guards on transitions");
        dataOutput.println("int (*gr"+id+"["+states.length+"]["+ii+
                "])(void)={{");
        first = true;
        counter = 0;
        for (int i = 0; i < states.length; i++) {
            if (first)
                first = false;
            else
                dataOutput.print("},{");
            counter = generateGuardTableLine(dataOutput, protocol, i, states, id, evs, counter);
        }
        dataOutput.println("}};");
        
        dataOutput.println();
        dataOutput.println("// Final states");
        dataOutput.print("int fn"+id+"[" + states.length + "]={");
        first = true;
        for (int i = 0; i < states.length; i++) {
            if (first)
                first = false;
            else
                dataOutput.print(",");
            if (states[i].isFinal()) {
                dataOutput.println("1");
            } else
                dataOutput.println("0");
        }
        dataOutput.println("};");
        dataOutput.println();
        
        // Synchronization offerFunctions
        generateOfferFunctions(dataOutput, states, id, multiSyncs, nevs, ii);
        dataOutput.println();
        
        output.println("c_decl{");
        output.println("\\#include \"component_data_"+id+".c\"");
        output.println("};");
        output.println();
        
        // Tracking state of local variables
        for (Map.Entry<String, BPVarDef> e : component.getVarDefs().entrySet()) {
            BPVarDef var = e.getValue();
            
            output.println("c_track \"&V" + id + "_" + e.getKey() +
                    "\" \"sizeof(&V" + id + "_" + e.getKey() + ")\"");
        }
        output.println();
        output.println("proctype c_" + id + "()");
        output.println("{");
        
        output.println();
        output.println("int st = " + startState + ";");
        output.println();
        
        String offer = "offer_" + id + "(Pc_"+id+"->st,tr" + id + ")";
        String unoffer = "unoffer_" + id + "(Pc_"+id+"->st,tr" + id + ")";
        
        output.println();
        output.println("c_code{ " + offer + ";};");
        output.println();
        
        output.println("do");
        
        for (int i = 0; i < et.numberOfEvents(); i++) {
            if (evs[i]){
                EBPEventTableImpl.AssignmentEvent assEvent;
                
                if (et.isAccept(i)) {
                    output.println("::accept(st,tr"+id+"[Pc_"+id+"->st][" + 
                            nevs[i] + "],gr"+id+"[Pc_"+id+"->st][" + 
                            nevs[i] + "]," + offer + "," + unoffer + 
                            "," + eventToBitVar(i) +
                            ",\"" + eventToStr(i) + "\");");
                } else if (et.isEmit(i)) {
                    output.println("::emit(st,tr"+id+"[Pc_"+id+"->st][" + 
                            nevs[i] + "],gr"+id+"[Pc_"+id+"->st][" + 
                            nevs[i] + "]," + offer + "," + unoffer + 
                            "," + eventToBitVar(i) + 
                            ");");
                } else if ( (assEvent = et.getAssignment(i)) != null ) {
                    output.println("::assign(st,tr"+id+"[Pc_"+id+"->st][" +
                            nevs[i] + "],gr"+id+"[Pc_"+id+"->st][" + 
                            nevs[i] + "]," + offer + "," + unoffer + 
                            ",V" + id + "_" +
                            assEvent.getVariable() + "," + assEvent.getValue() +
                            ");");
                } else if (et.isLambda(i)) {
                    output.println("::lambda(st,tr"+id+"[Pc_"+id+"->st][" + 
                            nevs[i] + "],gr"+id+"[Pc_"+id+"->st][" + 
                            nevs[i] + "]," + offer + "," + unoffer + 
                            ");");
                } else if (et.isMultisync(i)) {
                    Map<String, Integer> map = multiSyncs.get(i);
                    int idx = map.get(id);
                    
                    if (idx == 0 && map.size() == 1) {
                        // Single synchronisation
                        output.println("::synchro_single(st,tr"+id+"[Pc_"+id+"->st][" + 
                                nevs[i] + "],gr"+id+"[Pc_"+id+"->st][" + 
                                nevs[i] + "]," + offer + "," + unoffer + 
                                "," + eventToBitVar(i) +
                                ",\"" + eventToStr(i) + "\");");
                    } else if (idx == 0) {
                        // First synchronized
                        output.println("::synchro_first(st,tr"+id+"[Pc_"+id+"->st][" + 
                                nevs[i] + "],gr"+id+"[Pc_"+id+"->st][" + 
                                nevs[i] + "]," + offer + "," + unoffer + 
                                "," + eventToSyncCount(i) + "," + map.size() + 
                                "," + eventToBitVar(i) +
                                ",\"" + eventToStr(i) + "\");");
                    } else if (idx != map.size() - 1) {
                        // In the middle
                        output.println("::synchro_middle(st,tr"+id+"[Pc_"+id+"->st][" + 
                                nevs[i] + "],gr"+id+"[Pc_"+id+"->st][" + nevs[i] + 
                                "]," + offer + "," + unoffer + 
                                "," + eventToBitVar(i) + "," + idx + ");");
                    } else {
                        // Last synchronized
                        output.println("::synchro_last(st,tr"+id+"[Pc_"+id+"->st][" + 
                                nevs[i] + "],gr"+id+"[Pc_"+id+"->st][" + 
                                nevs[i] + "]," + offer + "," + unoffer + 
                                "," + eventToBitVar(i) + ");");
                    }
                    map.size();
                    
                } else {
                    // Unsuported event type
                    assert false;
                }
            }
        }
        output.println("::final(fn"+id+"[Pc_"+id+"->st])");
        // generateAssignments(output, assEvents, id, assEvents.size());
        output.println("od;");
        
        output.println(EXIT_STATE_LABEL + ":");
        output.println("skip;");
        output.println("}");
    }
    
    /**
     * Generates information about types (#define).
     */
    private void generateTypes(PrintStream output,
            BPAbstractComponent component, String protocolId) {
        output.println("// Type definitions");
        for (Map.Entry<String, BPTypeDef> e : component.getTypeDefs().entrySet()) {
            String type = e.getKey();
            int i = 0;
            boolean first = true;
            
            output.println("typedef enum {");
            
            for (String val : e.getValue().getValues()) {
                // NOTE: Types should not overlap in their domains
                if (first) {
                    first = false;
                    output.print("\t");
                } else {
                    output.print(", ");
                }
                output.print(val);
            }
            output.println();
            output.println("} T" + protocolId + "_" +
                    type + ";");
        }
    }
    
    /**
     * Generates offer synchronization functions
     */
    private void generateOfferFunctions(PrintStream output, 
            SimpleState[] stateTable, String protocolId, 
            Map<Integer, Map<String, Integer>> multiSync, int nevs[], 
            int levs) {

        output.println("// Synchronization offer functions");

        // Offer part
        output.println("inline void " + SYNC_OFFER_PREFIX +
                "_" + protocolId + " (int st, int tr["+stateTable.length+"]["+levs+"]) {");
        for (int ev : multiSync.keySet()) {
            if (!multiSync.get(ev).containsKey(protocolId))
                continue;
            //output.println("\t" + eventToSyncCount(ev) + " += tr[st*" + levs + "+" + nevs[ev] + "] != -1;");
            output.println("\t" + eventToSyncCount(ev) + " += tr[st][" + nevs[ev] + "] != -1;");
        }
        output.println("}");
        
        // Unoffer part
        output.println("inline void " + SYNC_UNOFFER_PREFIX +
                "_" + protocolId + " (int st, int tr["+stateTable.length+"]["+levs+"]) {");
        for (int ev : multiSync.keySet()) {
            if (!multiSync.get(ev).containsKey(protocolId))
                continue;
            //output.println("\t" + eventToSyncCount(ev) + " -= tr[st*" + levs + "+" + nevs[ev] + "] != -1;");
            output.println("\t" + eventToSyncCount(ev) + " -= tr[st][" + nevs[ev] + "] != -1;");
        }
        output.println("}");
    }
    
    /**
     * Generates the assignment "if" promela statement for a given set of
     * assignments.
     */
    private void generateAssignments(PrintStream output,
            Set<Integer> assEvents, String id, int assCount) {
        int i = 0;
        
        if (assEvents.size() == 0)
            return;
        
        output.println("::atomic{");
        output.println("\tc_expr{ass"+id+"[Pc_"+id+"->st]};");
        output.println("\tif");
        
        for (int event : assEvents) {
            // TODO: get the assignment code
            String doItCode = "TODO";
            
            output.println("::d_step{c_expr{ass_tr"+id+"[Pc_"+id+"->st*" +
                    assCount + "+" + i + "]!=-1;" + doItCode + ";" +
                    "st=c_expr{ass_tr"+id+"[Pc_"+id+"->st*" +
                    assCount + "+" + i + "]}}");
            i++;
        }
        
        output.println("\tfi;");
        output.println("}");
    }
    
    /**
     * Generates information about local variables.
     */
    private void generateVariables(PrintStream output,
            BPAbstractComponent component, String protocolId) {
        output.println("// Definition of local variables");
        
        for (Map.Entry<String, BPVarDef> e : component.getVarDefs().entrySet()) {
            BPVarDef var = e.getValue();
            
            output.println("T" + protocolId + "_" +
                    var.getType() + " V" + protocolId + "_" + e.getKey() +
                    " = " + var.getVal() + ";");
        }
    }
    
    /**
     * Generates a single state of the automaton (that is label and transitions
     * using if and other macros).
     */
    private void generateState(PrintStream output, ProtocolKeeper protocol,
            int state, SimpleState[] stateTable, String protocolId, boolean evs[]) {
        SimpleState s = stateTable[state];
        
        int e = s.nextEdge(State.BEFORE_FIRST_EDGE);
        int ee = 0;
        boolean first = true;
        
        while (e != State.AFTER_LAST_EDGE) {
            while (ee < e) {
                if (evs[ee]) {
                    if (first) {
                        first = false;
                        output.print("-1");
                    } else {
                        output.print(",-1");
                    }
                }
                ee++;
            }
            if (first)
                first = false;
            else
                output.print(",");
            output.print(s.getEdgeTarget(e));
            e = s.nextEdge(e);
            ee++;
        }
        while (ee < et.numberOfEvents()) {
            if (evs[ee]) {
                if (first) {
                    first = false;
                    output.print("-1");
                } else {
                    output.print(",-1");
                }
            }
            ee++;
        }
        
        output.println();
    }
    
    
    /**
     * Generates a single state of the automaton (that is label and transitions
     * using if and other macros).
     */
    private void generateAssignmentTableLine(PrintStream output,
            ProtocolKeeper protocol, int state, SimpleState[] stateTable,
            String protocolId, Set<Integer> assEvents) {
        SimpleState s = stateTable[state];
        
        int e = s.nextEdge(State.BEFORE_FIRST_EDGE);
        int ee = 0;
        boolean first = true;
        
        while (e != State.AFTER_LAST_EDGE) {
            while (ee < e) {
                if (assEvents.contains(ee)) {
                    if (first) {
                        first = false;
                        output.println("-1");
                    } else {
                        output.print(",-1");
                    }
                }
                ee++;
            }
            if (first)
                first = false;
            else
                output.print(",");
            output.println(s.getEdgeTarget(e));
            e = s.nextEdge(e);
            ee++;
        }
        while (ee < et.numberOfEvents()) {
            if (assEvents.contains(ee)) {
                if (first) {
                    first = false;
                    output.println("-1");
                } else {
                    output.print(",-1");
                }
            }
            ee++;
        }
    }
    
    /**
     * Generates a guard functions for transitions for a single state of the
     * automaton (that is label and transitions using if and other macros).
     */
    private int generateGuardFunctions(PrintStream output, ProtocolKeeper protocol,
            int state, SimpleState[] stateTable, String protocolId,
            boolean evs[], int counter) {
        
        SimpleState s = stateTable[state];
        int e = s.nextEdge(State.BEFORE_FIRST_EDGE);
        Guard2PromelaActions guard2prom = new Guard2PromelaActions();
        
        while (e != State.AFTER_LAST_EDGE) {
            Guard g = et.getGuard(e);
            if (!TrueGuard.instance().equals(g)) {
                // Generate the guard
                String gId = "G" + protocolId + "_" + String.valueOf(counter);
                
                // Generate the function code
                guard2prom.generateGuard(output, g, gId, protocolId);
                
                counter++;
            }
            e = s.nextEdge(e);
        }
        return counter;
    }
    
    /**
     * Generates a guard table line for transitions for a single state of the automaton
     * (that is label and transitions using if and other macros).
     */
    private int generateGuardTableLine(PrintStream output, ProtocolKeeper protocol,
            int state, SimpleState[] stateTable, String protocolId,
            boolean evs[], int counter) {
        SimpleState s = stateTable[state];
        
        int e = s.nextEdge(State.BEFORE_FIRST_EDGE);
        int ee = 0;
        boolean first = true;
        
        while (e != State.AFTER_LAST_EDGE) {
            while (ee < e) {
                if (evs[ee]) {
                    if (first) {
                        first = false;
                        output.print("0");
                    } else {
                        output.print(",0");
                    }
                }
                ee++;
            }
            if (first)
                first = false;
            else
                output.print(",");
            
            Guard g = et.getGuard(e);
            if (!TrueGuard.instance().equals(g)) {
                // Generate the guard
                String gId = "G" + protocolId + "_" + String.valueOf(counter);
                output.print(gId);
                counter++;
            } else {
                output.print("0");
            }
            
            e = s.nextEdge(e);
            ee++;
        }
        while (ee < et.numberOfEvents()) {
            if (evs[ee]) {
                if (first) {
                    first = false;
                    output.print("0");
                } else {
                    output.print(",0");
                }
            }
            ee++;
        }
        output.println();
        
        return counter;
    }
    
    /**
     * Generates accept transition identified by the given event.
     */
    private void generateFinalState(PrintStream output, String protocolId) {
        output.println("::" + FINAL_STATE_SIGNAL + "(" +
                EXIT_STATE_LABEL + ",fn" + protocolId + ");");
    }
    
    /**
     * Converts an event identification to String representing its Promela
     * bit variable.
     */
    private String eventToBitVar(int event) {
        
        if (et.isResp(event)) {
            return EVENT_BIT_VAR_PREFIX + EVENT_RESPONSE_PREFIX + et.getMethodCode(event);
        } else if (et.isRq(event)) {
            return EVENT_BIT_VAR_PREFIX + EVENT_REQUEST_PREFIX + et.getMethodCode(event);
        } else if (et.isMultisync(event)) {
            return EVENT_BIT_VAR_PREFIX + EVENT_MULTISYNC_PREFIX + et.getEventString(event).substring(1);
        }
        assert false;
        return null;
    }
    
    /**
     * Converts a multisync. event identification to String representing its 
     * Promela offer counter.
     */
    private String eventToSyncCount(int event) {
        
        if (et.isMultisync(event)) {
            return EVENT_COUNTER_VAR_PREFIX + EVENT_MULTISYNC_PREFIX + et.getEventString(event).substring(1);
        }
        assert false;
        return null;
    }
    
    /**
     * Converts an event identification to String method signature.
     */
    private String eventToStr(int event) {
        if (et.isCommunication(event)) {
            MethodSignature ms = et.getMethodEvent(event).getMethodSignature();
            String ret = ms.getIface() +
                    "." + ms.getMethod() +
                    "(";

            boolean first = true;
            for (Parameter p : ms.getParameters()) {
                if (first) {
                    first = false;
                } else {
                    ret += ", ";
                }
                ret += ((ActualParameter) p).getValue();
            }
            ret += ")" + (et.isRq(event) ? "^" : "$");

            return ret;
        } else {
            return et.getEventString(event);
        }
    }
    
    /**
     * Converts a state identification to String representing its Promela
     * label.
     */
    private String stateToLabel(int state) {
        return STATE_LABEL_PREFIX + state;
    }
    
    /**
     * Transform the preprocessed (i.e. instantiated) protocols into minimal
     * deterministic automata.
     *
     * @param architecture A preprocessed architecture whose protocols should
     *                      be minimalized in-place
     */
    private void prepareAutomata(PreprocessedArchitecture architecture) {
        int i = 1;
        for (BPComponent component : architecture.getComponents()) {
            Logger.getLogger("EBPTransformation").info("Minimalization started: " + i + "-th protocol");
            component.getProtocolKeeper().performTransformationChain(
                    Logger.getLogger("EBPTransformation"),
                    new EBPCreateUndetFSMActions(et));
            Logger.getLogger("EBPTransformation").info("Minimalization finished successfully!");
            i++;
        }
    }
    
    
    /**
     * The event table
     */
    EBPEventTableImpl et;
    
}
